@@ -5,20 +5,23 @@
 #
 # flake8: noqa: E501
 import base64
+import sys
 import importlib
 import datetime
 import json
-from decimal import Decimal
-
 from tina4_python import Debug, Constant
+from tina4_python.Constant import TINA4_LOG_ERROR
 from tina4_python.DatabaseResult import DatabaseResult
 
 
 class Database:
     SQLITE = "sqlite3"
     FIREBIRD = "firebird.driver"
+    FIREBIRD_INSTALL = "pip install firebird-driver or poetry add firebird-driver"
     MYSQL = "mysql.connector"
-    POSTGRES = "postgres"
+    MYSQL_INSTALL = "pip install mysql-connector-python or poetry add mysql-connector-python"
+    POSTGRES = "psycopg2"
+    POSTGRES_INSTALL = "pip install psycopg2-binary or poetry add psycopg2-binary"
 
     def __init__(self, _connection_string, _username="", _password=""):
         """
@@ -27,9 +30,22 @@ class Database:
         """
         # split out the connection string
         # driver:host/port:schema/path
-
         params = _connection_string.split(":", 1)
-        self.database_module = importlib.import_module(params[0])
+
+        try:
+            self.database_module = importlib.import_module(params[0])
+        except Exception as e:
+            install_message = "What driver are we working with?"
+            if params[0] == Database.SQLITE:
+                install_message = "Your python is missing the sqlite3 module, please reinstall or update"
+            elif params[0] == Database.MYSQL:
+                install_message = "Your python is missing the mysql module, please install with "+Database.MYSQL_INSTALL
+            elif params[0] == Database.POSTGRES:
+                install_message = "Your python is missing the postgres module, please install with "+Database.POSTGRES_INSTALL
+            elif params[0] == Database.FIREBIRD:
+                install_message = "Your python is missing the firebird module, please install with "+Database.FIREBIRD_INSTALL
+
+            sys.exit("Could not load database driver for "+params[0]+"\n"+install_message)
 
         self.database_engine = params[0]
         self.database_path = params[1]
@@ -40,7 +56,40 @@ class Database:
             self.dba = self.database_module.connect(self.database_path)
             self.port = None
             self.host = None
-        elif self.database_engine == self.FIREBIRD:
+
+            # we need to register data adapters for sqlite3 due to deprecations in python3.12
+            def adapt_date_iso(val):
+                """Adapt datetime.date to ISO 8601 date."""
+                return val.isoformat()
+
+            def adapt_datetime_iso(val):
+                """Adapt datetime.datetime to timezone-naive ISO 8601 date."""
+                return val.isoformat()
+
+            def adapt_datetime_epoch(val):
+                """Adapt datetime.datetime to Unix timestamp."""
+                return int(val.timestamp())
+
+            self.database_module.register_adapter(datetime.date, adapt_date_iso)
+            self.database_module.register_adapter(datetime.datetime, adapt_datetime_iso)
+            self.database_module.register_adapter(datetime.datetime, adapt_datetime_epoch)
+
+            def convert_date(val):
+                """Convert ISO 8601 date to datetime.date object."""
+                return datetime.date.fromisoformat(val.decode())
+
+            def convert_datetime(val):
+                """Convert ISO 8601 datetime to datetime.datetime object."""
+                return datetime.datetime.fromisoformat(val.decode())
+
+            def convert_timestamp(val):
+                """Convert Unix epoch timestamp to datetime.datetime object."""
+                return datetime.datetime.fromtimestamp(int(val))
+
+            self.database_module.register_converter("date", convert_date)
+            self.database_module.register_converter("datetime", convert_datetime)
+            self.database_module.register_converter("timestamp", convert_timestamp)
+        else:
             # <host>/<port>:<file>
             temp_params = self.database_path.split(":", 1)
             host_port = temp_params[0].split("/", 1)
@@ -52,33 +101,85 @@ class Database:
 
             self.database_path = temp_params[1]
 
-            self.dba = self.database_module.connect(
-                self.host + "/" + str(self.port) + ":" + self.database_path,
-                user=self.username,
-                password=self.password
-            )
-        else:
-            temp_params = self.database_path.split(":", 1)
-            host_port = temp_params[0].split("/", 1)
-            self.host = host_port[0]
-            if len(host_port) > 1:
-                self.port = int(host_port[1])
+            if self.database_engine == self.FIREBIRD:
+                self.dba = self.database_module.connect(
+                    self.host + "/" + str(self.port) + ":" + self.database_path,
+                    user=self.username,
+                    password=self.password
+                )
+            elif self.database_engine == self.MYSQL:
+                self.dba = self.database_module.connect(
+                    database=self.database_path,
+                    port=self.port,
+                    host=self.host,
+                    user=self.username,
+                    password=self.password
+                )
+            elif self.database_engine == self.POSTGRES:
+                self.dba = self.database_module.connect(
+                    dbname=self.database_path,
+                    port=self.port,
+                    host=self.host,
+                    user=self.username,
+                    password=self.password
+                )
+
             else:
-                self.port = 0
-
-            self.database_path = temp_params[1]
-
-            self.dba = self.database_module.connect(
-                database=self.database_path,
-                port=self.port,
-                host=self.host,
-                user=self.username,
-                password=self.password
-            )
+                sys.exit("Could not load database driver for "+params[0])
 
         Debug("DATABASE:", self.database_module, self.host, self.port, self.database_path, self.username,
               Constant.TINA4_LOG_DEBUG)
 
+    def table_exists(self, table_name):
+        """
+        Checks if a table exists in the database
+        :param str table_name: Name of the table
+        :return: bool : True if table exists, else False
+        """
+
+        sql = ""
+        if self.database_engine == self.SQLITE:
+            sql = "SELECT count(*) as count_table FROM sqlite_master WHERE type='table' AND name='"+table_name+"'"
+        elif self.database_engine == self.MYSQL:
+            sql = "SELECT count(*) as count_table FROM information_schema.tables WHERE table_schema = '"+self.database_path+"' AND table_name = '"+table_name+"'"
+        elif self.database_engine == self.POSTGRES:
+            sql = """SELECT count(*) as count_table FROM pg_catalog.pg_class c
+                        JOIN   pg_catalog.pg_namespace n ON n.oid = c.relnamespace
+                        WHERE  c.relname = '"""+table_name+"""'
+                        AND    c.relkind = 'r'        """
+        elif self.database_engine == self.FIREBIRD:
+            sql = "SELECT count(*) as count_table FROM RDB$RELATIONS WHERE RDB$RELATION_NAME = upper('"+table_name+"')"
+        else:
+            return False
+
+        record = self.fetch_one(sql)
+        if record:
+            if record["count_table"] > 0:
+                return True
+            else:
+                return False
+        else:
+            return False
+
+    def get_next_id(self, table_name, column_name="id"):
+        """
+        Gets the next id using max method in sql for databases which don't have good sequences
+        :param str table_name: Name of the table
+        :param str column_name: Name of the column in that table to increment
+        :return: int : The next id in the sequence
+        """
+        try:
+            sql = "select max(" + column_name + ") as \"max_id\" from " + table_name
+            record = self.fetch_one(sql)
+            if record["max_id"] is None:
+                record = {"max_id": 0}
+
+            next_id = int(record["max_id"]) + 1
+            return next_id
+        except Exception as e:
+            Debug("Get next id", str(e), TINA4_LOG_ERROR)
+            return None
+
     def database_exists(self, database_name):
 
         return True
@@ -108,6 +209,11 @@ class Database:
         return DatabaseResult(rows, columns, None)
 
     def is_json(self, myjson):
+        """
+        Checks if a JSON string is valid
+        :param myjson:
+        :return:
+        """
         try:
             json.loads(myjson)
         except Exception as e:
@@ -117,17 +223,19 @@ class Database:
     def fetch(self, sql, params=[], limit=10, skip=0):
         """
         Fetch records based on a sql statement
-        :param sql:
-        :param params:
-        :param limit:
-        :param skip:
-        :return:
+        :param str sql: A plain SQL statement or one with params in it designated by ?
+        :param list params: A list of params in order of precedence
+        :param int limit: Number of records to fetch
+        :param int skip: Offset of records to skip
+        :return: DatabaseResult
         """
         # modify the select statement for limit and skip
         if self.database_engine == self.FIREBIRD:
             sql = f"select first {limit} skip {skip} * from ({sql}) as t"
-        elif self.database_engine == self.SQLITE:
+        elif self.database_engine == self.SQLITE or self.database_engine == self.MYSQL:
             sql = f"select * from ({sql}) as t limit {skip},{limit}"
+        elif self.database_engine == self.POSTGRES:
+            sql = f"select * from ({sql}) as t limit {limit} offset {skip}"
         else:
             sql = f"select * from ({sql}) as t limit {skip},{limit}"
 
@@ -137,16 +245,16 @@ class Database:
             cursor.execute(sql, params)
             return self.get_database_result(cursor)
         except Exception as e:
-            Debug("FETCH ERROR:", sql, "params", params, "limit", limit, "skip", skip, Constant.TINA4_LOG_DEBUG)
+            Debug("FETCH ERROR:",  sql, str(e), "params", params, "limit", limit, "skip", skip, Constant.TINA4_LOG_DEBUG)
             return DatabaseResult(None, [], str(e))
 
     def fetch_one(self, sql, params=[], skip=0):
         """
-        Fetch a single record based on a sql statement
-        :param sql:
-        :param params:
-        :param skip:
-        :return:
+        Fetch a single record based on a sql statement, take note that BLOB and byte record data is converted into base64 automatically
+        :param str sql: A plain SQL statement or one with params in it designated by ?
+        :param list params: A list of params in order of precedence
+        :param int skip: Offset of records to skip
+        :return: dict : A dictionary containing the single record
         """
         # Calling the fetch method with limit as 1 and returning the result
         sql = self.parse_place_holders(sql)
@@ -154,6 +262,8 @@ class Database:
         if record.error is None and record.count == 1:
             data = {}
             for key in record.records[0]:
+                if isinstance(record.records[0][key], (datetime.date, datetime.datetime)):
+                    data[key] = record.records[0][key].isoformat()
                 if isinstance(record.records[0][key], bytes):
                     data[key] = base64.b64encode(record.records[0][key]).decode('utf-8')
                 else:
@@ -172,18 +282,17 @@ class Database:
         :param sql:
         :return:
         """
-
-        if self.database_engine == self.MYSQL:
+        if self.database_engine == self.MYSQL or self.database_engine == self.POSTGRES:
             return sql.replace("?", "%s")
         else:
             return sql.replace("%s", "?")
 
-    def execute(self, sql, params=()):
+    def execute(self, sql, params=[]):
         """
         Execute a query based on a sql statement
-        :param sql:
-        :param params:
-        :return:
+        :param str sql: A plain SQL statement or one with params in it designated by ?
+        :param list params: A list of params in order of precedence
+        :return: DatabaseResult
         """
         sql = self.parse_place_holders(sql)
         cursor = self.dba.cursor()
@@ -204,12 +313,12 @@ class Database:
             # Return the error in the result
             return DatabaseResult(None, [], str(e))
 
-    def execute_many(self, sql, params=()):
+    def execute_many(self, sql, params=[]):
         """
         Execute a query based on a single sql statement with a different number of params
-        :param sql:
-        :param params:
-        :return:
+        :param sql: A plain SQL statement or one with params in it designated by ?
+        :param params: A list of params in order of precedence
+        :return: DatabaseResult
         """
         sql = self.parse_place_holders(sql)
         cursor = self.dba.cursor()
@@ -224,13 +333,19 @@ class Database:
             return DatabaseResult(None, [], str(e))
 
     def start_transaction(self):
+        """
+        Starts a transaction
+        :return:
+        """
         try:
-            if self.database_engine in (self.SQLITE, self.POSTGRES):
+            if self.database_engine == self.SQLITE:
                 self.dba.execute("BEGIN TRANSACTION")
             elif self.database_engine == self.FIREBIRD:
                 self.dba.begin()
             elif self.database_engine == self.MYSQL:
                 self.dba.start_transaction()
+            elif self.database_engine == self.POSTGRES:
+                self.dba.rollback() #start fresh
             else:
                 Debug("START TRANSACTION ERROR:", "Database engine unrecognised/not supported",
                       Constant.TINA4_LOG_ERROR)
@@ -238,18 +353,30 @@ class Database:
             Debug("START TRANSACTION ERROR:", str(e), Constant.TINA4_LOG_ERROR)
 
     def commit(self):
+        """
+        Commit transaction
+        :return:
+        """
         try:
             self.dba.commit()
         except Exception as e:
             Debug("COMMIT TRANSACTION ERROR:", str(e), Constant.TINA4_LOG_ERROR)
 
     def rollback(self):
+        """
+        Rollback transaction
+        :return:
+        """
         try:
             self.dba.rollback()
         except Exception as e:
             Debug("ROLLBACK TRANSACTION ERROR:", str(e), Constant.TINA4_LOG_ERROR)
 
     def close(self):
+        """
+        Close database connection
+        :return:
+        """
         try:
             self.dba.close()
         except Exception as e:
@@ -261,7 +388,6 @@ class Database:
         :param record:
         :return:
         """
-
         for key in record:
             if isinstance(record[key], list) or isinstance(record[key], dict):
                 record[key] = json.dumps(record[key])
@@ -270,9 +396,10 @@ class Database:
     def insert(self, table_name, data, primary_key="id"):
         """
         Insert data based on table name and data provided - single or multiple records
-        :param primary_key:
-        :param table_name:
-        :param data:
+        :param str table_name: Name of table
+        :param None data: List or Dictionary containing the data to be inserted
+        :param str primary_key: The name of the primary key of the table
+        :return:
         """
         if isinstance(data, dict):
             data = [data]
@@ -289,10 +416,12 @@ class Database:
 
             sql = f"INSERT INTO {table_name} ({columns}) VALUES ({placeholders})"
 
-            if self.database_engine == self.FIREBIRD or self.database_engine == self.SQLITE:
+            if self.database_engine == self.FIREBIRD or self.database_engine == self.SQLITE or self.database_engine == self.POSTGRES:
                 sql += f" returning ({primary_key})"
 
             records = DatabaseResult()
+
+            result = None
             for record in data:
                 record = self.sanitize(record)
                 result = self.execute(sql, list(record.values()))
@@ -308,12 +437,10 @@ class Database:
     def delete(self, table_name, filter=None):
         """
         Delete data based on table name and filter provided - single or multiple filters
-        :param table_name:
-        :param records:
-        :param primary_key:
-        :param filter:
+        :param str table_name: Name of table
+        :param str filter: Expression for deleting records
+        :return: bool
         """
-
         if self.database_engine in (self.SQLITE, self.FIREBIRD):
             placeholder = "?"
         elif self.database_engine in (self.MYSQL, self.POSTGRES):
@@ -328,6 +455,8 @@ class Database:
 
             # Updating multiple records - records passed in is a list
             if isinstance(filter, list):
+                sql = ""
+                result = None
                 for record in filter:
                     pk_value = []
                     condition_records = []
@@ -352,14 +481,14 @@ class Database:
                     Debug("DELETE ERROR:", sql, result.error, Constant.TINA4_LOG_ERROR)
                     return False
 
-    def update(self, table_name, records, primary_key="id"):
+    def update(self, table_name, data, primary_key="id"):
         """
         Update data based on table name and record/primary key provided - single or multiple records
-        :param table_name:
-        :param records:
-        :param primary_key:
+        :param str table_name: Name of table
+        :param None data: List or Dictionary containing the data to be inserted
+        :param str primary_key: The name of the primary key of the table
+        :return:
         """
-
         if self.database_engine in (self.SQLITE, self.FIREBIRD):
             placeholder = "?"
         elif self.database_engine in (self.MYSQL, self.POSTGRES):
@@ -367,14 +496,16 @@ class Database:
         else:
             placeholder = "?"
 
-        if records is not None:
+        if data is not None:
             # Updating a single record - record passed in is a dictionary
-            if isinstance(records, dict):
-                records = [records]
+            if isinstance(data, dict):
+                data = [data]
 
             # Updating multiple records - records passed in is a list
-            if isinstance(records, list):
-                for record in records:
+            if isinstance(data, list):
+                sql = ""
+                result = None
+                for record in data:
                     pk_value = None
                     condition_records = ""
                     set_clause_list = []
